/*
* Ping Sensor wiring
*   VCC connection of the sensor attached to +5V
*   GND connection of the sensor attached to ground
*   TRIG connection of the sensor attached to digital pin 2
*   ECHO connection of the sensor attached to digital pin 8
*/
  
#include <Servo.h>
#include "Hardware.c"
#include "Filter.c"
#include "Fire.c"
#include "PitchMath.c"


/*******************************************************************************
 * Global Variables
 ******************************************************************************/
//SERVOS
  Servo yaw;  //Can't be defined in Hardware.c because it requires the Servo library, which is C++
  Servo pitch;
  
//The scanning arrays
  int yawArray[180]; 
    //array that holds the values at each angle horizontally.  Will be overwritten with each use.
  int pitchArray[180]; 
    //array that holds the values at each angle vertically when we do our vertical scan.
    
//How many vertical positions at which we make a horizontal scan
  int yawCenters[3];        //array that holds the horizontal center for each vertical positon we scan at
  
//Distance and scanning set by our range switch settings
  int lowValue;        //Used in the filter, set in the setDistance method
  int highValue;       //Used in the filter, set in the setDistance method
  int pitchOffset;      //How far in between vertical passes

  
int ammo = 6;             //Number of shots left
int yawBeginAngle = YAW_BASE_ANGLE;
int yawEndAngle = YAW_MAX_ANGLE;


/*******************************************************************************
 * Arduino Program Functions
 ******************************************************************************/

void setup() {
  //Initialize pins and hardware
  initHardware(); //Defined in Hardware.c
  //Initialize servos
  yaw.attach(basePin); 
  pitch.attach(topPin); 
  
  //Initialize arrays
  for(int i = 0; i < 180; i++){
    yawArray[i] = -1;
  }
  for(int i = 0; i < (sizeof(yawCenters)/sizeof(yawCenters[0])); i++){
    yawCenters[i] = -1;
  }
  for(int i = 0; i < (sizeof(pitchArray)/sizeof(pitchArray[0])); i++){
    pitchArray[i] = -1;
  }
  
}

void loop()
{
  int inches; //Distance from sonar at each angle
  int range; //The optimal range set by the switches 
  //TARGETING COORDINATES
    int averageYaw; //The calculated best horizontal position.
    int centerPitch; //The center value from our vertical scan
  
  //Initialize servo position, wait for button press.
  initServos();
  Serial.println("Test Begins");
  while(digitalRead(startButton) == HIGH){
      digitalWrite(errorLED, HIGH); //Blink the LED so we know it's ready
      delay(100);
      digitalWrite(errorLED, LOW);
      delay(100);
  }
  digitalWrite(errorLED, LOW);//Turn off LED when we start
  
  //Read range settings from switches, set distance variables and window angles
  //range = readDistance();
  setDistance(0); //Sets high/low values and yaw begin/end values
  Serial.print("RANGE: Low value: ");
  Serial.print(lowValue);
  Serial.print(". High value ");
  Serial.println(highValue);

  /* BEGIN FIELD SCANNING---------------------------------------------------------------
   * Algorithm for scanning the field
   * 
   * For each vertical position
   *   Perform a horizontal scan, saving values in yawArray[]
   *   Filter yawArray
   *   Calculate center value and save in yawCenters[]
   * Calculate the average of the yawCenters and save value as averageYaw
   * Go to average yaw and perform a vertical scan, saving values in pitchArray[]
   * Filter pitchArray
   * Calculate center value of pitchArray and save value as centerPitch
   */
   for(int height = 0; height < (sizeof(yawCenters)/sizeof(yawCenters[0])); height++)
   {
     int pitchPos = PITCH_BASE_ANGLE + (pitchOffset * height);
     //Position pitch servo and wait for movement to complete
     target(YAW_BASE_ANGLE, pitchPos);
       //Beginning with PITCH_BASE_ANGLE, every iteration through the loop will add pitchOffset to the angle
     delay(100);
     
     // HORIZONTAL SCAN **************************************************************
     for(int yawPos = yawBeginAngle; yawPos < yawEndAngle; yawPos++){
       //position yaw servo and wait for movement to complete
       target(yawPos, pitchPos);
       delay(100);

       inches = ping();
       //Serial.println(inches);
       
       //Write distance to array
       yawArray[yawPos] = inches;
     }//end horizontal scan
     
    //Output to serial THIS IS TESTING
    printArray(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) ); //TODO Remove

    //FILTER THE HORIZONTAL ARRAY*******************************************************
    //filterData(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])), highValue, lowValue );
    fieldFilterData(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) );
        //Output to serial THIS IS TESTING
        printArray(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) ); //TODO Remove
        
    filterCalibrationTargets(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) );
        printArray(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) ); //TODO Remove
        
    smoothFilterData(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) );
       
    //Output to serial
    printArray(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) );
       
    //Calculate the center of the target and save to yawCenters
    yawCenters[height] = findCenter(yawArray, (sizeof(yawArray)/sizeof(yawArray[0])) );
    Serial.print("Saving target center from horizontal scan: ");
    Serial.println(yawCenters[height]);
   }//end for each yawCenter[height]
   
   Serial.println("Calculate the average of the yaw centers");

   averageYaw = selectYawCenter(yawCenters, (sizeof(yawCenters)/sizeof(yawCenters[0])) ); 
     
   //Go to (averageYaw, PITCH_BASE_ANGLE) and perform a vertical scan
   target(averageYaw, PITCH_BASE_ANGLE);
   for(int i = PITCH_BASE_ANGLE; i < PITCH_MAX_ANGLE; i++){
     //position yaw servo and wait for movement to complete
     target(averageYaw, i);
     delay(100);
     //Serial.println(i);

     inches = ping();
       
     //Save distance in array
     pitchArray[i] = inches;
   }
     
   //Filter the vertical array
   filterData(pitchArray, (sizeof(pitchArray)/sizeof(pitchArray[0])), highValue, lowValue);
   //smoothFilterData(pitchArray, (sizeof(pitchArray)/sizeof(pitchArray[0])));
     
   //Calculate the vertical center
   centerPitch = findCenter(pitchArray, (sizeof(pitchArray)/sizeof(pitchArray[0])) );
   //COMPENSATE/BIAS THE VERTICAL CENTER
   if(centerPitch > 18) centerPitch = centerPitch - 12;
   
   //Print vertical array to serial
   printArray(pitchArray, sizeof(pitchArray)/sizeof(pitchArray[0]) );
   
   //END FIELD SCANNING------------------------------------------------------------------------------------------
 
   //Output target coordinates
   Serial.print("Target center calculated at (");
   Serial.print(averageYaw);
   Serial.print(", ");
   Serial.print(centerPitch);
   Serial.println(") degrees.");
   
   //Calculate angle to shoot, target it
   target(averageYaw, centerPitch);
   delay(500);
   int finalDistance = ping();
   int adjustedPitch = adjustPitch(centerPitch, finalDistance);
   Serial.print("Adjusted pitch: ");
   Serial.print(adjustedPitch);
   Serial.print(" degrees at range (inches) of ");
   Serial.println(finalDistance);
   
   //Decide how to shoot and SHOOT!
   target(averageYaw, adjustedPitch);
   delay(500);//Allow movement to finish
   
   if(finalDistance >= 72 - 6)       shootSix(averageYaw, adjustedPitch);        //From 6 feet onward, 
   else if(finalDistance >= 48 - 6)  shootVertical(averageYaw, adjustedPitch);   //From 4 to six feet
   else                              shootVertical(averageYaw, adjustedPitch);      //Less than 4 feet
   shootAll(averageYaw, adjustedPitch); //In case of misfire, clear the gun by shooting everything
  
   
}//End loop()


/*******************************************************************************
 * Hardware and Servo Functions
 ******************************************************************************/

/*
 * Initialize the pins for the dip switches, servos, and LEDS.
 * Start the serial connection.
 * Can't be moved to Hardware.c file because INPUT_PULLUP and OUTPUT
 *  constants won't be recognized.
 */
void initHardware()
{
  //set the disctance switches
  pinMode(binDigOne, INPUT_PULLUP);
  pinMode(binDigTwo, INPUT_PULLUP);
  pinMode(binDigThr, INPUT_PULLUP);
  pinMode(binDigFou, INPUT_PULLUP);
  pinMode(startButton, INPUT_PULLUP);
  

  pinMode(errorLED, OUTPUT); 
  pinMode(firingPin, OUTPUT);
  Serial.begin(9600);
}

/*
 * Dip switches used to set the range to expect a target at.
 * Reads the four switches as binary digits, and converts them to 
 *   a little-endian integer.  Returns this int, which will be the
 *   range to target.
 *
 * Can't be moved to Hardware.c file because HIGH and LOW are 
 *  undeclared.
 */
int readDistance()
{
  //read switches
  int output = 0;
  if(digitalRead(binDigFou) == HIGH)
  { 
    output = 0;
  }
  else if(digitalRead(binDigFou) == LOW)
  {
    output = 8;
  }
  if(digitalRead(binDigThr) == HIGH)
  { 
    output = output + 0;
  }
  else if(digitalRead(binDigThr) == LOW)
  {
    output = output + 4;
  }
  if(digitalRead(binDigTwo) == HIGH)
  { 
    output = output + 0;
  }
  else if(digitalRead(binDigTwo) == LOW)
  {
    output = output + 2;
  }
  if(digitalRead(binDigOne) == HIGH)
  { 
    output = output + 0;
  }
  else if(digitalRead(binDigOne) == LOW)
  {
    output = output + 1;
  }
  Serial.print("readDistance reads ");
  Serial.print(output);
  Serial.println(" feet from dip switches");
  return output;
}//End readDistance()

//Convert microsecond delay from sonar to distance in inches.
int microsecondsToInches(long microseconds)
{
  return (int) microseconds / 74 / 2;
}
  
/*
 * Position servos in their base positions.
 *
 * Can't be moved to hardware.c file because we're using
 *   C++ servo objects.
 */
void initServos()
{
  target(YAW_BASE_ANGLE, PITCH_BASE_ANGLE);
  delay(1000);
}

/*******************************************************************************
 * Program Supporting Functions
 ******************************************************************************/

//Print the array to Serial on a single line.
void printArray(int array[], int length)
{
  Serial.println("\nArray:");
  for(int i = 0; i < length; i++){
    Serial.print(array[i]);
    Serial.print(",");
  }
  Serial.println();
}

/* 
 * Find the center of the largest string of values
 */
int findCenter(int array[], int length)
{
  int beginAngle = -1;
  int targetLength = 0;
  int offset = -1;
  int targetCenter = -1;
  
  //The best target length (for comparison)
    int bestLength = 0;
  //The angle of the best target center.  -1 if there are no targets in this array (used in 2D targeting)
    int bestLengthLoc = -1;
  
  for(int i = 0; i < length; i++){
    //Count targets only for nonzero values, and end before the last value in the array
    //"If a nonzero value AND the difference of the value and the next is less than 12 AND we're not at the end of the array"
    if( (array[i] > 0) && absolute(array[i] - array[i+1]) < 12 && (i < length - 1) ){
      if(beginAngle == -1){ //BEGINNING OF A TARGET
        beginAngle = i;
        //Serial.print("findCenters() identified a beginAngle: ");
        //Serial.println(i);
      }
      targetLength++; //ANY VALUE IN A "TARGET"
    }
    else{ //Value is zero or we're at the last index in the array: calculate the center
      if(beginAngle != -1){ //If we have been counting a target
        offset = targetLength / 2;
        targetCenter = beginAngle + offset;
          //Serial.print("findCenter() identified a target center: ");
          //Serial.println(targetCenter);
        //If this is the largest target, save the center
        if(targetLength > bestLength){
          bestLength = targetLength; //Save this target's length
          bestLengthLoc = targetCenter;//Save this target's center position
            //Serial.print("findCenter() saved a new bestLength: ");
            //Serial.print(bestLength);
            //Serial.print(" at location: ");
            //Serial.println(bestLengthLoc);
        }
      }
             
      //Reset variables
      beginAngle = -1;
      offset = -1;
      targetLength = 0;
      targetCenter = -1;
    }
  }
  
  Serial.print("findCenter() is about to return a bestLengthLoc of: ");
  Serial.println(bestLengthLoc);
  if(bestLengthLoc <= 60 && bestLengthLoc >= 30){
    bestLengthLoc -= 5;
    Serial.println("findCenter() BIASING RIGHT 5 DEGREES");
  }
  if(bestLengthLoc >= 120 && bestLengthLoc <= 150){
    bestLengthLoc += 5;
    Serial.println("findCenter() BIASING LEFT 5 DEGREES");
  }
  return bestLengthLoc;
}

/* Sets the global variables lowValue, highValue, yawBeginAngle, yawEndAngle, and pitchAngle
 * based on the given range
 * @param distance the range setting from the dip switches (in whole feet)
 * is passed in.
 * @return Whether the range setting is 1 - 10 feet or special zero setting.
 */
boolean setDistance(int distance)
{
  int inchesDistance;
  int angleOffset; //The angle to skip on the far edge of each side per distance
 
 if(distance == 0){
   lowValue = 6;
   highValue = 126;
 }
 else{ //For any range setting besides zero.
  //set Low and High Values
  inchesDistance = distance * 12;
  lowValue = inchesDistance - 6;
  highValue = inchesDistance + 6;
  
  //Set begin and end Angles
  switch(distance){
    case 1:
    case 2:
    case 3:
    case 4: //all fall through to five feet
    case 5:
      angleOffset = 0;
      break;
    case 6:
      angleOffset = 33;
      break;
    case 7:
      angleOffset = 44;
      break;
    case 8:
      angleOffset = 51;
      break;
    case 9:
      angleOffset = 56;
      break;
    case 10:
      angleOffset = 60;
      break;
    case 11:
      angleOffset = 63;
      break;
    case 12:
      angleOffset = 65;
      break;
    case 13:
      angleOffset = 67;
      break;
    case 14:
      angleOffset = 69;
      break;
    case 15:
      angleOffset = 70;
      break;
    default:
      angleOffset = 1;
      break;
   }
  yawBeginAngle = 0 + angleOffset;
  yawEndAngle = 180 - angleOffset;
 }//end else
 
  pitchOffset = 14; //Sets the global pitchOffset variable to 14 for now.  //TODO change based on distance.
  return distance < 10;
}

//Returns inches to obstacle.  Ping with the sonar, average return times, and convert the average to inches.
int ping()
{ 
  //Pulse and listen
  pinMode(trigPin, OUTPUT);		//Should this be moved to a helper method
  digitalWrite(trigPin, LOW);		// such as inches = ping() ?
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH);
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);  
  pinMode(echoPin, INPUT);
  long duration = pulseIn(echoPin, HIGH);
  
  return microsecondsToInches(duration);
}

/*
 * Selects the yawCenter that is closest to the center (90degrees) of the field.
 */
int selectYawCenter(int array[], int length)
{
  int bestOffset = 180; //The lowest distance from 90degrees
    //no result will be farther away from center than "backwards"
  int bestCenter = 90; //The yawCenter angle that has the lowest distance from 90deg.
    //if the yawCenters Array is full of -1s, try shooting straight forward
    
  for(int i = 0; i < length; i++){
    int offset = abs(array[i] - 90); //distance this yawCenter is from 90deg
    if(offset < bestOffset){ //if closer than the previous best
      bestOffset = offset;
      bestCenter = array[i];
    } 
  }
  
  Serial.print("returning best yawCenter of ");
  Serial.println(bestCenter);
  return bestCenter;
}

/******************************************************************************
 * Targeting and Firing
 * ***************************************************************************/

/*
 * Targets a set of coordinates, bounds-checking for servo safety
 * Paramenters: x-coordinate (yaw), y-coordinate (pitch)
 */
void target(int xCoord, int yCoord)
{
  //Bounds check
  if(xCoord > 180) xCoord = 180;
  if(xCoord < YAW_BASE_ANGLE) xCoord = YAW_BASE_ANGLE;
  if(yCoord > PITCH_MAX_ANGLE) yCoord = PITCH_MAX_ANGLE;
  if(yCoord < PITCH_BASE_ANGLE) yCoord = PITCH_BASE_ANGLE;
  
  yaw.write(xCoord);
  pitch.write(yCoord);
}

/*
 * See Fire.c for function comments
 */
void shootInvCross(int xCoord, int yCoord)
{
    //Shot one: center
    target(xCoord, yCoord);
    delay(500);
    fireShot();
  
    //Shot two: right 1 degree
    target(xCoord - SHOOT_CROSS_SPACE, yCoord);
    delay(500);
    fireShot();
  
    //Shot three: up one degree
    target(xCoord, yCoord + SHOOT_CROSS_SPACE);
    delay(500);
    fireShot();
  
    //Shot four: left one degree
    target(xCoord + SHOOT_CROSS_SPACE, yCoord);
    delay(500);
    fireShot();
  
    //Shot five: down one degree
    target(xCoord, yCoord - SHOOT_CROSS_SPACE);
    delay(500);
    fireShot();
  
    //Shot six, up two degrees
    target(xCoord, yCoord + 2 * SHOOT_CROSS_SPACE);
    delay(500);
    fireShot();
}

/*
 * Targets a set of coordinates and shoots a volley in a pattern at and
 * around those coordinates in a right-side-up cross pattern
 */
void shootCross(int xCoord, int yCoord)
{
    //Shot one: center
    target(xCoord, yCoord);
    delay(500);
    fireShot();
  
    //Shot two: right 1 degree
    target(xCoord - SHOOT_CROSS_SPACE, yCoord);
    delay(500);
    fireShot();
  
    //Shot three: up one degree
    target(xCoord, yCoord + SHOOT_CROSS_SPACE);
    delay(500);
    fireShot();
  
    //Shot four: left one degree
    target(xCoord + SHOOT_CROSS_SPACE, yCoord);
    delay(500);
    fireShot();
  
    //Shot five: down one degree
    target(xCoord, yCoord - SHOOT_CROSS_SPACE);
    delay(500);
    fireShot();
  
    //Shot six, down two degrees
    target(xCoord, yCoord - 2 * SHOOT_CROSS_SPACE);
    delay(500);
    fireShot();
}

/*
 * Targets a set of coordinates and shoots a volley in a vertical line
 * 	pattern across the target
 */
void shootVertical(int xCoord, int yCoord)
{
  //Shot one: at target coordinates
  target(xCoord, yCoord);
  delay(500);
  fireShot();
  
  //Shot two: up two degres
  target(xCoord, yCoord + 2);
  delay(500);
  fireShot();
  
  //Shot three: up four degrees
  target(xCoord, yCoord + 4);
  delay(500);
  fireShot();
  
  //Shot four: down two degrees
  target(xCoord, yCoord - 3);
  delay(500);
  fireShot();
  
  //Shot five: down four degrees
  target(xCoord, yCoord - 6);
  delay(500);
  fireShot();
  
  //Shot six: at target coordinates again (in case shot one missed)
  target(xCoord, yCoord - 8);
  delay(500);
  fireShot();
}

/*
 * Targets a set of coordinates and shoots all six darts
 * 	at a controlled speed.
 */
void shootSix(int xCoord, int yCoord)
{
  target(xCoord, yCoord);
  for(int i = 0; i < 12; i++){ //Run twice (2x the six darts) just in case
    delay(500);
    fireShot();
  }
}

/*
 * Fire a shot from the dart gun.
 * THIS FUNCTION IS CALLED BY shoot() in the file "Fire.c"
 * Has to be defined here because it uses the HIGH and LOW constants.
 */
void fireShot()
{
  digitalWrite(firingPin, HIGH);
  delay(200); //How long to power the launcher to fire a dart.
  digitalWrite(firingPin, LOW);
  delay(250); //Safety delay, wait for the shot to get out
  Serial.println("FIRING!");
}

/*
 * Rapidfires ALL SHOTS from the dart gun.
 * (As a contingency in case of misfires)
 * Has to be defined here because it uses the HIGH and LOW constants.
 */
void shootAll(int xCoord, int yCoord)
{
  target(xCoord, yCoord);
  delay(500);
  
  digitalWrite(firingPin, HIGH);
  delay(1200); //How long to power the launcher to fire a dart.
  digitalWrite(firingPin, LOW);
  delay(250); //Safety delay, wait for the shot to get out
  Serial.println("FIRING ALL!");
}

int absolute(int number)
{
  if(number < 0){
    number = number * -1;
  }
  
  return number;
}
