/* PitchMath.c
 * 
 * Math function to adjust pitch angle targeting to get the right trajectory
 * 	to hit a target.
 * Modified Feb 19 2014
 */


/* Direct line angles to a point 1.5 feet above level at i+1 feet away
 * Servo-sonar angle will be 19 degrees higher than the real-life angle starting at level.
 * Calculated by arcTangent(18 in / distance-to-target-in-inches) + 19
 */
const int HIGH_TARGET_LOCATION[10] = {75, 56, 46, 40, 33, 30, 29, 30, 28, 28};

//What angle to shoot to hit a target at +1.5 feet above the table at i+1 feet away
const int HIGH_ANGLE_PITCHES[10] = {54, 40, 34, 37, 37, 40, 45, 45, 45, 45};
  //REMEMBER THAT ARRAYS START AT ZERO AND TARGETS START AT ONE FOOT
//What angle to shoot to hit a target at level with the table at i+1 feet away
const int LOW_ANGLE_PITCHES[10] = {7, 12, 15, 17, 25, 28, 35, 48, 48, 48};


/******************************************************************************
 * Prototypes and Comments
 *****************************************************************************/


/*
 * Returns the pitch angle required to hit a target with the trajectory arc
 * 	of our darts.
 * @param inches the distance to the target
 * @param pitchToTarget the angle above center where the target center is located
 * RATIO USED: PTT / HSA == TPO / TPW
 *   PTT: pitchToTarget
 *   HSA: highSightAngle
 *   TPO: trajectoryPitchOffset
 *   TPW: trajectoryPitchWindow
 */
int adjustPitch(int inches, int pitchToTarget);


/******************************************************************************
 * Function Definitions
 *****************************************************************************/


int adjustPitch(int inches, int pitchToTarget)
{
  int trajectoryPitch; //The pitch to shoot so the trajectory arc will hit center
  
  int index = ( (int) ( ( (double)inches + 6.0 ) / 12.0)  ) - 1; //Convert to nearest foot starting at 6inches, subtract one to make an array index.
  if(index < 0) index = 0;
  
  int highSightAngle = HIGH_TARGET_LOCATION[index]; //The highest angle at which we would look at the center of a target at this distance
  int trajectoryPitchOffset; //How far above the "Low angle" to shoot (up to a maximum of the "High Angle"
  int trajectoryPitchWindow = HIGH_ANGLE_PITCHES[index] - LOW_ANGLE_PITCHES[index]; //The size of the space between low and high.
  
  //Clip pitchToTarget to angle boundaries.
  if(pitchToTarget > highSightAngle){
    pitchToTarget = highSightAngle;
  }
  
  //Calculate the trajectoryPitchOffset with the ratio (see function comments above)
  trajectoryPitchOffset = (int) ( ( (double)pitchToTarget / (double)highSightAngle ) * (double)trajectoryPitchWindow );
  
  //Add the offset to the base angle
  trajectoryPitch = LOW_ANGLE_PITCHES[index] + trajectoryPitchOffset;
  
  return trajectoryPitch - 5;//COMPENSATION
}
